# -*- coding: utf-8 -*-

import json

from elasticsearch_dsl import Document, Keyword

__all__ = ("BaseDocument", "Config", "IndexPattern", "Visualization", "Dashboard")


class BaseDocument(Document):
    type = Keyword()

    # List of json attributes.
    json_attrs = []

    class Meta:
        doc_type = "doc"

    class Index:
        # We use elasticsearch_dsl 6.3 behaviour here as we don't know in advance
        # the name of the kibana index
        name = "*"

    def __init__(self, **kwargs):
        super().__init__(type=self._type, **kwargs)
        self._json_attrs_cache = {}

    def __getattr__(self, key):
        if key in self.json_attrs:
            if key not in self._json_attrs_cache:
                self._json_attrs_cache[key] = json.loads(
                    self.to_dict()[self._type].get(key, "null")
                )
            return self._json_attrs_cache[key]
        return super().__getattr__(key)


class Config(BaseDocument):
    _type = "config"


class IndexPattern(BaseDocument):
    _type = "index-pattern"
    json_attrs = ["fields", "fieldFormatMap"]


class Search(BaseDocument):
    _type = "search"

    def index(self):
        """
        Returns the index-pattern associated to the visualization. Go through the
        search if needed.
        """
        search_source = self.search["kibanaSavedObjectMeta"]["searchSourceJSON"]
        key = json.loads(search_source).get("index")
        return IndexPattern.get(id=f"index-pattern:{key}", index=self.meta.index)


class Visualization(BaseDocument):
    _type = "visualization"
    json_attrs = ["visState", "uiStateJSON"]

    def related_search(self):
        """
        Returns the search associated to the visualization.

        An error is raised if the visualization is not associated to any search.
        """
        return Search.get(
            id=f"search:{self.visualization.savedSearchId}", index=self.meta.index
        )

    def index(self):
        """
        Returns the index-pattern associated to the visualization. Go through the
        search if needed.
        """
        if hasattr(self.visualization, "savedSearchId"):
            return self.related_search().index()
        search_source = self.visualization.kibanaSavedObjectMeta.searchSourceJSON
        key = json.loads(search_source).get("index")
        return IndexPattern.get(id=f"index-pattern:{key}", index=self.meta.index)

    def filters(self):
        """
        Returns the search filters
        :return elasticsearch_dsl.Q
        """
        from pybana import FilterTranslator

        search_source = self.visualization.kibanaSavedObjectMeta.searchSourceJSON
        filters = json.loads(search_source).get("filter", [])
        return FilterTranslator().translate(filters)


class Dashboard(BaseDocument):
    _type = "dashboard"
    json_attrs = ["panelsJSON", "optionsJSON"]

    def visualizations(self, missing="skip", using=None):
        """
        Does the join automatically by parsing panelsJSON.

        :param missing: Check https://elasticsearch-dsl.readthedocs.io/en/latest/api.html#elasticsearch_dsl.Document.mget
        :type string
        """
        panels = self.panelsJSON
        return (
            Visualization.mget(
                docs=["visualization:" + panel["id"] for panel in panels],
                index=self.meta.index,
                missing=missing,
                using=using,
            )
            if panels
            else []
        )



# -*- coding: utf-8 -*-

import elasticsearch_dsl
import hjson
import json

from pybana.translators.elastic.buckets import BucketTranslator, compute_auto_interval
from pybana.translators.elastic.metrics import MetricTranslator
from .filter import FilterTranslator

__all__ = ("ElasticTranslator", "FilterTranslator")


class ElasticTranslator:
    def translate_vega(self, visualization, scope):
        # TODO: handle magic keyword %timefilter%
        def replace_autointerval(node):
            if isinstance(node, list):
                return [replace_autointerval(child) for child in node]
            elif isinstance(node, dict):
                ret = {}
                for key, val in node.items():
                    if (
                        key == "interval"
                        and isinstance(val, dict)
                        and val.get("%autointerval%")
                    ):
                        ret[key] = compute_auto_interval("auto", scope.beg, scope.end)
                    else:
                        ret[key] = replace_autointerval(val)
                return ret
            return node

        spec = hjson.loads(visualization.visState["params"]["spec"])
        data = spec["data"] if isinstance(spec["data"], dict) else spec["data"][0]
        index = data["url"]["index"]
        body = replace_autointerval(data["url"]["body"])
        search = elasticsearch_dsl.Search(index=index).from_dict(body)
        if data["url"].get("%timefield%"):
            ts = data["url"]["%timefield%"]
            search = search.filter(
                "range",
                **{ts: {"gte": scope.beg.isoformat(), "lte": scope.end.isoformat()}}
            )
        return search

    def translate_legacy(self, visualization, scope):
        index_pattern = visualization.index()
        index = index_pattern["index-pattern"]["title"]
        ts = index_pattern["index-pattern"]["timeFieldName"]
        search = elasticsearch_dsl.Search(index=index).filter(
            "range",
            **{ts: {"gte": scope.beg.isoformat(), "lte": scope.end.isoformat()}}
        )
        state = json.loads(visualization.visualization["visState"])
        segment_aggs = [
            agg
            for agg in state["aggs"]
            if agg["schema"] in ("segment", "group", "split", "bucket")
        ]
        metric_aggs = [agg for agg in state["aggs"] if agg["schema"] in ("metric",)]
        proxy = search.aggs
        for agg in segment_aggs:
            proxy = BucketTranslator().translate(proxy, agg, state, scope)
        for agg in metric_aggs:
            MetricTranslator().translate(proxy, agg, state)
        search = search[:0]
        search = search.filter(visualization.filters())
        return search

    def translate(self, visualization, scope):
        """
        Transform a kibana visualization object into an elasticsearch_dsl Search.

        :param elasticsearch_dsl.Document visualization: Visualization fetched from a kibana index.
        :param Scope scope: Scope to use for data fetching.
        """
        if visualization.visState["type"] == "vega":
            return self.translate_vega(visualization, scope)
        else:
            return self.translate_legacy(visualization, scope)



# -*- coding: utf-8 -*-

import json
import os

from elasticsearch import NotFoundError
import elasticsearch_dsl

from .models import Config, Dashboard, IndexPattern, Visualization, Search

__all__ = ("Kibana",)

DEFAULT_CONFIG = {
    "timepicker:timeDefaults": json.dumps(
        {"from": "now-7d", "to": "now", "mode": "quick"}
    ),
    "dateFormat:tz": "UTC",
    "state:storeInSessionStorage": True,
    "telemetry:optIn": False,
    "defaultIndex": None,
}


class Kibana:
    """
    Kibana client.
    """

    klasses = {
        "dashboard": Dashboard,
        "visualization": Visualization,
        "index-pattern": IndexPattern,
        "search": Search,
    }

    def __init__(self, index=".kibana"):
        """
        Initialize a client to kibana.

        :param index string: Index used by kibana (default: .kibana).
        """
        self._index = index

    def _search(self, type, using):
        klass = self.klasses.get(type)
        search = klass.search if klass else elasticsearch_dsl.Search
        return search(index=self._index, using=using)

    def _get(self, klass, id, using):
        ret = klass.get(index=self._index, id=id, using=using)
        return ret

    def objects(self, type, using=None):
        return self._search(type, using=using).filter("term", type=type)

    def config_id(self, using=None):
        elastic = elasticsearch_dsl.connections.get_connection(using or "default")
        return "config:%s" % elastic.info()["version"]["number"]

    def config(self, using=None):
        """
        Return the config associated to the current version of elastic
        """
        return self._get(Config, self.config_id(using), using=using)

    def init_index(self):
        """
        Create the elasticsearch index as kibana would do.
        """
        elastic = elasticsearch_dsl.connections.get_connection("default")
        mappingsfn = os.path.join(os.path.dirname(__file__), "mappings.json")
        suffix = 1
        while not elastic.indices.exists(self._index):
            index = f"{self._index}_{suffix}"
            if not elastic.indices.exists(index):
                with open(mappingsfn) as fd:
                    elastic.indices.create(index, body=json.load(fd))
                    elastic.indices.put_alias(index=index, name=self._index)
                break
            suffix += 1

    def init_config(self, using=None):
        """
        Create the config document that each kibana requires. This
        document stores all the settings such as timepicker defaults,
        date formats etc
        """
        try:
            self.config(using=using)
        except NotFoundError:
            Config(config=DEFAULT_CONFIG, meta={"id": self.config_id()}).save(
                index=self._index, refresh="wait_for", using=using
            )

    def index_patterns(self, using=None):
        """
        Return a Search to all the index-patterns.
        """
        return self.objects("index-pattern", using=using)

    def index_pattern(self, id, using=None):
        """
        Return a index-pattern identified by its identifier.
        """
        return self._get(
            self.klasses["index-pattern"], f"index-pattern:{id}", using=using
        )

    def searches(self, using=None):
        """
        Return a Search to all the index-patterns.
        """
        return self.objects("search", using=using)

    def search(self, id, using=None):
        """
        Return a index-pattern identified by its identifier.
        """
        return self._get(self.klasses["search"], f"search:{id}", using=using)

    def visualizations(self, using=None):
        """
        Return a Search to all the visualizations.
        """
        return self.objects("visualization", using=using)

    def visualization(self, id, using=None):
        """
        Return a visualization identified by its identifier.
        """
        return self._get(
            self.klasses["visualization"], f"visualization:{id}", using=using
        )

    def dashboards(self, using=None):
        """
        Return a Search to all the dashboards.
        """
        return self.objects("dashboard", using=using)

    def dashboard(self, id, using=None):
        """
        Return a dashboard identified by its identifier.
        """
        return self._get(self.klasses["dashboard"], f"dashboard:{id}", using=using)

    def update_or_create_default_index_pattern(self, index_pattern, using=None):
        """
        If config document does not have index pattern, associate the
        first index pattern found.
        """
        config = self.config(using)
        if not config.config.to_dict().get("defaultIndex"):
            config.config.defaultIndex = index_pattern.meta.id.split(":")[-1]
            config.save(refresh="wait_for")



# -*- coding: utf-8 -*-

import ast
import datetime
import elasticsearch
import elasticsearch_dsl
import json
import os
import pytest
import pytz
import sys

BASE_DIRECTORY = os.path.join(os.path.dirname(__file__), "..")  # NOQA
sys.path.insert(0, BASE_DIRECTORY)  # NOQA

from pybana import Scope, ElasticTranslator, Kibana, VegaRenderer, VegaTranslator
from pybana.translators.elastic.buckets import (
    format_from_interval,
    compute_auto_interval,
)

PYBANA_INDEX = ".kibana_pybana_test"
elastic = elasticsearch.Elasticsearch()
elasticsearch_dsl.connections.add_connection("default", elastic)


def load_fixtures(elastic, kibana, index):
    for realindex in elastic.indices.get(index, ignore_unavailable=True):
        elastic.indices.delete(realindex)
    datafn = os.path.join(BASE_DIRECTORY, "pybana/index.json")

    kibana.init_index()

    def actions():
        with open(datafn, "r") as fd:
            for line in fd:
                if not line:
                    continue
                action = json.loads(line)
                action["_index"] = index
                yield action

    elasticsearch.helpers.bulk(elastic, actions(), refresh="wait_for")


def load_data(elastic, index):
    ts = datetime.datetime(2019, 1, 1)
    if elastic.indices.exists(index):
        elastic.indices.delete(index)
    elastic.indices.create(
        index,
        body={
            "mappings": {
                "doc": {
                    "properties": {
                        "s": {"type": "keyword"},
                        "i": {"type": "integer"},
                        "f": {"type": "float"},
                        "ts": {"type": "date"},
                    }
                }
            }
        },
    )

    def actions():
        for i in range(100):
            yield {
                "_index": index,
                "_type": "doc",
                "_id": str(i),
                "_source": {
                    "ts": ts + datetime.timedelta(hours=i),
                    "f": float(i),
                    "i": i,
                    "s": chr(100 + (i % 10)),
                },
            }

    elasticsearch.helpers.bulk(elastic, actions(), refresh="wait_for")


def test_client():
    kibana = Kibana(PYBANA_INDEX)
    elastic.indices.delete(f"{PYBANA_INDEX}*")
    elastic.indices.create(f"{PYBANA_INDEX}_1")
    load_fixtures(elastic, kibana, PYBANA_INDEX)
    kibana.init_config()
    kibana.init_config()
    assert kibana.config()
    assert len(list(kibana.index_patterns())) == 1
    index_pattern = kibana.index_pattern("6c172f80-fb13-11e9-84e4-078763638bf3")
    index_pattern.fields
    index_pattern.fieldFormatMap
    kibana.update_or_create_default_index_pattern(index_pattern)
    kibana.update_or_create_default_index_pattern(index_pattern)
    visualizations = list(kibana.visualizations().scan())
    assert len(visualizations) == 25
    visualization = kibana.visualization("6eab7cb0-fb18-11e9-84e4-078763638bf3")
    visualization.visState
    visualization.uiStateJSON
    assert visualization.index().meta.id == index_pattern.meta.id
    dashboards = list(kibana.dashboards())
    assert len(dashboards) == 1
    dashboard = kibana.dashboard("f57a7160-fb18-11e9-84e4-078763638bf3")
    dashboard.panelsJSON
    dashboard.optionsJSON
    assert len(dashboard.visualizations()) == 2
    visualization = kibana.visualization("f4a09a00-fe77-11e9-8c18-250a1adff826")
    search = visualization.related_search()
    assert search.meta.id == "search:2139a4e0-fe77-11e9-833a-0fef2d7dd143"
    assert len(list(kibana.searches())) == 1
    search = kibana.search("2139a4e0-fe77-11e9-833a-0fef2d7dd143")
    assert visualization.index().meta.id == index_pattern.meta.id


def test_translators():
    kibana = Kibana(PYBANA_INDEX)
    load_fixtures(elastic, kibana, PYBANA_INDEX)
    load_data(elastic, "pybana")
    kibana.init_config()
    translator = ElasticTranslator()
    scope = Scope(
        datetime.datetime(2019, 1, 1, tzinfo=pytz.utc),
        datetime.datetime(2019, 1, 3, tzinfo=pytz.utc),
        pytz.utc,
        kibana.config(),
    )
    for visualization in kibana.visualizations().scan():
        if visualization.visState["type"] in ("histogram", "metric", "pie", "line"):
            search = translator.translate(visualization, scope)
            if visualization.meta.id.split(":")[-1] in (
                "695c02f0-fb1a-11e9-84e4-078763638bf3",
                "1c7226e0-ffd9-11e9-b6bd-4d907ad3c29d",
                "cdecdff0-ffd9-11e9-b6bd-4d907ad3c29d",
                "fa5fcfc0-ffd9-11e9-b6bd-4d907ad3c29d",
                "3e6c9e50-ffda-11e9-b6bd-4d907ad3c29d",
                "65881000-ffda-11e9-b6bd-4d907ad3c29d",
                "5fa0ea20-ffdc-11e9-b6bd-4d907ad3c29d",
                "86457e20-ffdc-11e9-b6bd-4d907ad3c29d",
                "ad4b9310-ffdc-11e9-b6bd-4d907ad3c29d",
                "e19d9640-ffdc-11e9-b6bd-4d907ad3c29d",
                "2fb77bc0-ffdd-11e9-b6bd-4d907ad3c29d",
                "9a4d3520-013f-11ea-b1ec-3910cd795dc1",
                "53b3da70-fbbc-11e9-84e4-078763638bf3",
                "c36b8b00-6f85-11ea-85b8-8f688e91da4a",
            ):
                response = search.execute()
                VegaTranslator().translate(visualization, response, scope)


def test_vega_visuzalization():
    kibana = Kibana(PYBANA_INDEX)
    load_fixtures(elastic, kibana, PYBANA_INDEX)
    load_data(elastic, "pybana")
    kibana.init_config()
    translator = ElasticTranslator()
    scope = Scope(
        datetime.datetime(2019, 1, 1, tzinfo=pytz.utc),
        datetime.datetime(2019, 1, 3, tzinfo=pytz.utc),
        pytz.utc,
        kibana.config(),
    )
    keys = [
        "37589ee0-70f3-11ea-9898-e1fb79cbf2bc",
        "da8f8510-7107-11ea-9898-e1fb79cbf2bc",
    ]
    for key in keys:
        visualization = kibana.visualization(key)
        search = translator.translate(visualization, scope)
        assert (
            search.to_dict()["aggs"]["category"]["date_histogram"]["interval"] == "1h"
        )

        response = search.execute()
        VegaTranslator().translate(visualization, response, scope)


def test_elastic_translator_helpers():
    assert format_from_interval("1y") == "yyyy"
    assert format_from_interval("1q") == "yyyy-MM"
    assert format_from_interval("1M") == "yyyy-MM"
    assert format_from_interval("1d") == "yyyy-MM-dd"
    assert format_from_interval("1h") == "yyyy-MM-dd'T'HH'h'"
    assert format_from_interval("1s") == "date_time"

    assert (
        compute_auto_interval("d", datetime.datetime.now(), datetime.datetime.now())
        == "1d"
    )

    def f(*args, **kwargs):
        delta = datetime.timedelta(*args, **kwargs)
        end = datetime.datetime.now()
        beg = end - delta
        return compute_auto_interval("auto", beg, end)

    assert f(days=800) == "30d"
    assert f(days=400) == "1w"
    assert f(days=50) == "1d"
    assert f(days=20) == "12h"
    assert f(days=5) == "3h"
    assert f(days=2) == "1h"
    assert f(days=1) == "30m"
    assert f(seconds=43000) == "10m"
    assert f(seconds=13000) == "5m"
    assert f(seconds=3600) == "1m"
    assert f(seconds=1200) == "30s"
    assert f(seconds=600) == "10s"
    assert f(seconds=240) == "5s"
    assert f(seconds=1) == "1s"


def test_vega_renderer():
    renderer = VegaRenderer()
    renderer.to_svg({"$schema": "https://vega.github.io/schema/vega/v5.json"})


def test_datasweet():
    import pybana.helpers.datasweet as ds

    assert ds.ds_avg(0, 1) == .5
    assert ds.ds_count(0, 1) == 2
    assert ds.ds_cusum(0, 1) == [0, 1]
    assert ds.ds_derivative(0, 1)[1] == 1
    assert ds.ds_max(0, 1) == 1
    assert ds.ds_min(0, 1) == 0
    assert ds.ds_next(0, 1)[0] == 1
    assert ds.ds_prev(0, 1)[1] == 0
    assert ds.ds_sum(0, 1) == 1

    assert ds.is_variable("agg1")
    assert not ds.is_variable("xagg1")

    assert (
        ds.datasweet_eval("avg(agg1, agg2) + 1", {"1": {"value": 0}, "2": {"value": 1}})
        == 1.5
    )

    assert ds.datasweet_eval("1 / 0", {}) is None

    with pytest.raises(ValueError):
        tree = ast.parse("x + 1", mode="eval")
        tree = ds.DatasweetTransformer().visit(tree)



# -*- coding: utf-8 -*-

import json


__all__ = ("ContextVisualization",)


class ContextVisualization:
    """
    Represent a visualization with a context.

    :param Visualization visualization: Visualization deserialized.
    :param config pybana.Config: Config of the kibana instance.
    """

    def __init__(self, visualization, config):
        self._state = visualization.visState.to_dict()
        self._ui_state = visualization.uiStateJSON.to_dict()
        self._config = config
        self.ui_colors = {
            **json.loads(
                self._config.config.to_dict().get("visualization:colorMapping", "{}")
            ),
            **self._ui_state.get("vis", {}).get("colors", {}),
        }

    def singleton(self):
        return all(map(lambda agg: agg["schema"] != "segment", self._state["aggs"]))

    def _aggs_by_type(self, typ):
        return [agg for agg in self._state["aggs"] if agg["schema"] == typ]

    def type(self):
        return self._state["type"]

    def series_params(self, agg):
        return [
            param
            for param in self._state["params"]["seriesParams"]
            if param["data"]["id"] == agg["id"]
        ][0]

    def bucket_aggs(self):
        """
        Return all the aggregations that generate bucketing.
        """
        return [
            agg
            for agg in self._state["aggs"]
            if agg["schema"] in ("segment", "group", "bucket")
        ]

    def segment_aggs(self):
        return self._aggs_by_type("segment")

    def metric_aggs(self):
        return self._aggs_by_type("metric")

    def metric_label(self, agg):
        if agg["params"].get("customLabel"):
            return agg["params"]["customLabel"]
        if self.type() in ("pie",):
            if agg["type"] == "count":
                return "Count"
            elif agg["type"] == "sum":
                return "Sum of %(field)s" % agg["params"]
            elif agg["type"] == "cardinality":
                return "Unique count of %(field)s" % agg["params"]
            raise NotImplementedError(
                "%(type)s for pie chart is not implemented" % agg
            )  # pragma: no cover
        elif self.type() in ("table", "metric"):
            if agg["type"] == "count":
                return "Count"
            return "%s - %s" % (agg["type"], agg["params"]["field"])
        return self.series_params(agg)["data"]["label"]

    def group_aggs(self):
        return self._aggs_by_type("group")

    def groups_side_by_side(self, ax):
        return len(self.group_aggs()) > 0 and not self.groups_stacked(ax)

    def groups_stacked(self, ax):
        return (
            any(
                [
                    param.get("mode") == "stacked"
                    for param in self._state["params"]["seriesParams"]
                    if param["valueAxis"] == ax["id"]
                ]
            )
            and len(self.group_aggs()) > 0
        )

    def metrics_side_by_side(self, ax):
        return len(
            [
                agg
                for agg in self.metric_aggs()
                if self.series_params(agg)["valueAxis"] == ax["id"]
            ]
        ) > 1 and not self.metrics_stacked(ax)

    def metrics_stacked(self, ax):
        aggs = [
            (agg, self.series_params(agg))
            for agg in self.metric_aggs()
            if self.series_params(agg)["valueAxis"] == ax["id"]
        ]

        return (
            any([param.get("mode") == "stacked" for agg, param in aggs])
            and len(aggs) > 1
        )

    def multi_axis(self):
        aggs = self.metric_aggs()
        axis = set([self.series_params(agg)["valueAxis"] for agg in aggs])
        return len(axis) > 1

    def stacked_applied(self, ax):
        return self.metrics_stacked(ax) or self.groups_stacked(ax)

    def valueax(self, axid):
        return [ax for ax in self.valueaxes() if ax["id"] == axid][0]

    def valueaxtype(self, ax):
        return [
            serie
            for serie in self._state["params"]["seriesParams"]
            if serie["valueAxis"] == ax["id"]
        ][0]["type"]

    def valueaxes(self):
        return self._state["params"].get("valueAxes", [])

    def y(self, ax):
        axid = ax["id"].split("-")[-1]
        return f"y{axid}"



# -*- coding: utf-8 -*-

import json
import os
import subprocess

__all__ = ("VegaRenderer",)
VEGA_BIN = os.path.join(os.path.dirname(__file__), "../../bin/vega-cli")


class VegaRenderer:
    """
    Renderer which takes in input a vega spec and returns the svg code
    """

    def to_svg(self, spec):
        p = subprocess.Popen(
            [VEGA_BIN],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return p.communicate(input=json.dumps(spec).encode())[0].decode()



# -*- coding: utf-8 -*-

from elasticsearch_dsl import Q


class BaseFilterTranslator:
    def estype(self, filtr):
        return self.type

    def params(self, filtr):
        return filtr["meta"]["params"]

    def translate(self, filtr):
        return Q(self.estype(filtr), **{filtr["meta"]["key"]: self.params(filtr)})


class ExistsFilterTranslator(BaseFilterTranslator):
    type = "exists"

    def translate(self, filtr):
        return Q("exists", field=filtr["meta"]["key"])


class PhraseFilterTranslator(BaseFilterTranslator):
    type = "phrase"

    def estype(self, filtr):
        return "match_phrase"

    def params(self, filtr):
        return {"query": filtr["meta"]["value"]}


class PhrasesFilterTranslator(PhraseFilterTranslator):
    type = "phrases"

    def translate(self, filtr):
        q = None
        for param in filtr["meta"]["params"]:
            q1 = Q("match_phrase", **{filtr["meta"]["key"]: param})
            q = q1 if q is None else q | q1
        return q


class RangeFilterTranslator(BaseFilterTranslator):
    type = "range"


FILTER_TRANSLATORS = {
    filtr.type: filtr
    for filtr in (
        ExistsFilterTranslator,
        PhraseFilterTranslator,
        PhrasesFilterTranslator,
        RangeFilterTranslator,
    )
}


class FilterTranslator:
    def translate(self, filters):
        q = Q()
        for filtr in filters:
            if filtr["meta"]["disabled"]:
                continue
            translator = FILTER_TRANSLATORS[filtr["meta"]["type"]]()
            q1 = translator.translate(filtr)
            q &= ~q1 if filtr["meta"]["negate"] else q1
        return q



# -*- coding: utf-8 -*-

from pybana.helpers.datasweet import datasweet_eval

__all__ = ("VEGA_METRICS",)


class BaseMetric:
    def contribute(self, agg, bucket, response):
        return (bucket or response["aggregations"])[agg["id"]]["value"]


class CountMetric(BaseMetric):
    aggtype = "count"

    def contribute(self, agg, bucket, response):
        return (
            bucket["doc_count"]
            if bucket and "doc_count" in bucket
            else response["hits"]["total"]
        )


class AverageMetric(BaseMetric):
    aggtype = "avg"


class MedianMetric(BaseMetric):
    aggtype = "median"

    def contribute(self, agg, bucket, response):
        return (bucket or response["aggregations"])[agg["id"]]["values"]["50.0"]


class StdDevMetric(BaseMetric):
    aggtype = "std_dev"

    def contribute(self, agg, bucket, response):
        return (bucket or response["aggregations"])[agg["id"]]["std_deviation"]


class MinMetric(BaseMetric):
    aggtype = "min"


class MaxMetric(BaseMetric):
    aggtype = "max"


class SumMetric(BaseMetric):
    aggtype = "sum"


class CardinalityMetric(BaseMetric):
    aggtype = "cardinality"


class DatasweetMetric(BaseMetric):
    aggtype = "datasweet_formula"

    def contribute(self, agg, bucket, response):
        return datasweet_eval(agg["params"]["formula"], bucket)


VEGA_METRICS = {
    metric.aggtype: metric
    for metric in [
        AverageMetric,
        CardinalityMetric,
        CountMetric,
        DatasweetMetric,
        MaxMetric,
        MedianMetric,
        MinMetric,
        StdDevMetric,
        SumMetric,
    ]
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

with open("requirements.txt") as requirements_file:
    requirements = [req.strip("\n") for req in requirements_file.readlines()]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest"]
setup(
    author="Guillaume Thomas",
    author_email="guillaume.thomas@inuse.eu",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Python client for kibana. Provide ORM & vega rendering of visualizations",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="pybana",
    name="pybana",
    packages=find_packages(include=["src/pybana"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/optimdata/pybana",
    version="0.5.0",
    zip_safe=False,
)



# -*- coding: utf-8 -*-

import hjson

from pybana.helpers import percentage

from .constants import (
    KIBANA_SEED_COLORS,
    DEFAULT_WIDTH,
    DEFAULT_PIE_WIDTH,
    DEFAULT_HEIGHT,
    DEFAULT_PADDING,
)
from .metrics import VEGA_METRICS
from .visualization import ContextVisualization

__all__ = ("VegaTranslator",)


class VegaTranslator:
    def conf(self, state):
        return {
            "$schema": "https://vega.github.io/schema/vega/v5.json",
            "width": DEFAULT_PIE_WIDTH if state.type() == "pie" else DEFAULT_WIDTH,
            "height": DEFAULT_HEIGHT,
            "padding": DEFAULT_PADDING,
        }

    def data(self, conf, state, response):
        if state.type() == "pie":
            conf = self.data_pie(conf, state, response)
        else:
            conf = self.data_line_bar(conf, state, response)
        return conf

    def data_pie(self, conf, state, response):
        conf["data"] = [
            {
                "name": "table",
                "values": [],
                "transform": [{"type": "pie", "field": "y"}],
            }
        ]
        # In case of a pie, there is only one metric agg
        metric_agg = state.metric_aggs()[0]
        metric = VEGA_METRICS[metric_agg["type"]]()
        if state.singleton():
            conf["data"][0]["values"].append(
                {
                    "y": metric.contribute(metric_agg, None, response),
                    "metric": state.metric_label(metric_agg),
                    "group": "all",
                }
            )
        else:

            def rec(segment_it):
                """
                TODO: Implement recursion.
                """
                segment_agg = state.segment_aggs()[segment_it]
                buckets = response.aggregations.to_dict()[segment_agg["id"]]["buckets"]
                sumbuckets = sum(
                    [
                        metric.contribute(metric_agg, bucket, response)
                        for bucket in buckets
                    ]
                )
                for bucket in buckets:
                    y = metric.contribute(metric_agg, bucket, response)
                    label = state.metric_label(metric_agg)
                    group = bucket["key"]
                    ratio = percentage(y, sumbuckets)
                    conf["data"][0]["values"].append(
                        {
                            "y": y,
                            "metric": label,
                            "segment": segment_it,
                            "group": group,
                            "tooltip": {group: f"{y}/{sumbuckets} ({ratio:.2f}%)"},
                            "label": "%s (%.2f%%)" % (group, ratio),
                        }
                    )

            rec(0)
        return conf

    def _iter_response(self, node, bucket_aggs, it, point, state, response):
        """
        Iterate through response and yield each point.

        A point is the value of a metric at a given bucketing. It has the following
        properties:
        - x: It's the value of the segment agg (there's should be only one)
        - group: A string which concatenate all the group keys
        - y: The value of the metric
        - metric: The name of the metric
        - axis: The axis on which the point should be displayed

        :param node dict: The node of the response. Should start at `response.aggregations`.
        :param bucket_aggs list: List of bucket aggs
        :param it: Iterator on bucket_aggs.
        :param point: The point currently being filled.
        :param state: Visualization state.
        :param response: Elasticsearch response.
        """
        if it == len(bucket_aggs):
            point["group"] = " - ".join(
                map(str, filter(bool, point.setdefault("groups", [])))
            )
            for m, metric_agg in enumerate(state.metric_aggs()):
                if metric_agg.get("hidden"):
                    continue
                metric = VEGA_METRICS[metric_agg["type"]]()
                y = metric.contribute(metric_agg, node, response)
                childpoint = point.copy()
                # handling case where no bucket aggs
                childpoint.setdefault("x", "all")
                childpoint.pop("groups")
                childpoint.update(
                    {"y": y, "m": m, "metric": state.metric_label(metric_agg)}
                )
                if "seriesParams" in state._state["params"]:
                    series_params = state.series_params(metric_agg)
                    ax = state.valueax(series_params["valueAxis"])
                    childpoint.update(
                        {state.y(ax): y, "axis": series_params["valueAxis"]}
                    )
                tooltip = {"x": childpoint["x"], childpoint["metric"]: y}
                if childpoint["group"]:
                    tooltip["group"] = childpoint["group"]
                childpoint["tooltip"] = tooltip
                yield childpoint
            return
        agg = bucket_aggs[it]
        aggnode = node[agg["id"]]
        for child in aggnode["buckets"]:
            childpoint = point.copy()
            key = child.get("key_as_string") or child.get("key")
            if agg["schema"] == "segment":
                childpoint["x"] = key
            else:
                childpoint.setdefault("groups", []).append(key)
            for obj in self._iter_response(
                child, bucket_aggs, it + 1, childpoint, state, response
            ):
                yield obj

    def data_line_bar(self, conf, state, response):
        data = {"name": "table", "values": []}
        for ax in state.valueaxes():
            if state.groups_stacked(ax):
                data["transform"] = [
                    {
                        "type": "stack",
                        "groupby": ["x"],
                        "field": state.y(ax),
                        "as": [state.y(ax) + "|0", state.y(ax) + "|1"],
                        "offset": "normalize"
                        if ax["scale"]["mode"] == "percentage"
                        else "zero",
                    }
                ]
            elif state.metrics_stacked(ax):
                data["transform"] = [
                    {
                        "type": "stack",
                        "groupby": ["x"],
                        "field": state.y(ax),
                        "as": [state.y(ax) + "|0", state.y(ax) + "|1"],
                        "sort": {"field": "metric"},
                        "offset": "normalize"
                        if ax["scale"]["mode"] == "percentage"
                        else "zero",
                    }
                ]

        for item in self._iter_response(
            response.aggregations.to_dict(), state.bucket_aggs(), 0, {}, state, response
        ):
            data["values"].append(item)

        for ax in state.valueaxes():
            if state.stacked_applied(ax):
                for row in data["values"]:
                    if row.get(state.y(ax)) is None:
                        row[state.y(ax)] = 0

        conf["data"] = [data]
        return conf

    def _scale_x(self, state):
        return {
            "name": "xscale",
            "type": "band" if state.type() == "histogram" else "point",
            "domain": {"data": "table", "field": "x"},
            "range": "width",
            "padding": 0.05,
            "round": True,
        }

    def _scale_axis(self, state):
        return {
            "name": "axiscolor",
            "type": "band",
            "domain": {"data": "table", "field": "axis"},
            "range": "category",
        }

    def _scales_metric(self, state, conf):
        if state.type() == "pie":
            return
        scheme = []
        domain = []
        for a, agg in enumerate(state.metric_aggs()):
            label = state.series_params(agg)["data"]["label"]
            ax = state.valueax(state.series_params(agg)["valueAxis"])
            # Let's hide legend for values that are always equal to 0 when they
            # are stacked. This hack is usefull when there are many metrics and
            # only few of them with data.
            if state.metrics_stacked(ax):
                label = state.metric_label(agg)
                if all(
                    [
                        row["y"] == 0
                        for row in conf["data"][0]["values"]
                        if row["metric"] == label
                    ]
                ):
                    continue

            domain.append(label)
            scheme.append(
                state.ui_colors.get(
                    label, KIBANA_SEED_COLORS[a % len(KIBANA_SEED_COLORS)]
                )
            )

        yield {
            "name": "metriccolor",
            "type": "ordinal",
            "range": scheme,
            "domain": domain,
        }

    def _scale_group(self, state, data):
        scheme = []
        domain = []
        groups = set()
        for point in data[0]["values"]:
            group = point["group"]
            if group not in groups:
                domain.append(group)
                color = state.ui_colors.get(
                    group, KIBANA_SEED_COLORS[len(groups) % len(KIBANA_SEED_COLORS)]
                )
                scheme.append(color)
                groups.add(group)

        return {
            "name": "groupcolor",
            "type": "ordinal",
            "range": scheme,
            "domain": domain,
        }

    def _scales_y(self, state):
        for ax in state.valueaxes():
            if "min" in ax["scale"] and "max" in ax["scale"]:
                domain = [ax["scale"]["min"], ax["scale"]["max"]]
            else:
                domain = {
                    "data": "table",
                    "field": state.y(ax) + "|1"
                    if state.stacked_applied(ax)
                    else state.y(ax),
                }
            yield (
                {"name": ax["id"], "domain": domain, "nice": True, "range": "height"}
            )

    def scales(self, conf, state):
        conf["scales"] = [
            self._scale_x(state),
            *self._scales_y(state),
            self._scale_axis(state),
            *self._scales_metric(state, conf),
            self._scale_group(state, conf["data"]),
        ]

        return conf

    def axes(self, conf, state):
        if state.type() in ("line", "histogram"):
            # TODO: handle more that 1 axe
            conf["axes"] = []
            categoryax = state._state["params"]["categoryAxes"][0]

            if categoryax["show"]:
                ax = {
                    "orient": categoryax["position"],
                    "scale": "xscale",
                    "labelOverlap": True,
                }
                if categoryax["labels"].get("rotate", 0) != 0:
                    ax.update(
                        {
                            "labelAngle": 360 - categoryax["labels"]["rotate"],
                            "labelBaseline": "middle",
                            "labelAlign": "right",
                            "labelLimit": 1000,
                        }
                    )
                conf["axes"].append(ax)
            for ax in state.valueaxes():
                if ax["show"]:
                    axconf = {
                        "orient": ax["position"],
                        "scale": ax["id"],
                        "title": ax["title"]["text"],
                    }
                    if ax["scale"]["mode"] == "percentage":
                        axconf["format"] = ".0%"
                    conf["axes"].append(axconf)
        return conf

    def legends(self, conf, state):
        if state.type() == "pie":
            conf = self.legends_pie(conf, state)
        else:
            conf = self.legends_line_bar(conf, state)
        return conf

    def legends_pie(self, conf, state):
        conf["legends"] = [
            {
                "fill": "groupcolor",
                "title": "Series",
                "orient": state._state["params"]["legendPosition"],
                # TODO: offset should be dynamic given the data to prevent label/legend overlapping
                "offset": 150,
            }
        ]
        return conf

    def legends_line_bar(self, conf, state):
        if len(state.group_aggs()):
            conf["legends"] = [
                {
                    "fill": "groupcolor",
                    "title": "Series",
                    "columns": 1
                    if state._state["params"]["legendPosition"] in ("left", "right")
                    else 10,
                    "orient": state._state["params"]["legendPosition"],
                }
            ]
        else:
            conf["legends"] = [
                {
                    "fill": "metriccolor",
                    "title": "Series",
                    "columns": 1
                    if state._state["params"]["legendPosition"] in ("left", "right")
                    else 10,
                    "orient": state._state["params"]["legendPosition"],
                }
            ]
        return conf

    def marks(self, conf, state, response):
        if state.type() == "pie":
            conf = self.marks_pie(conf, state, response)
        elif state.type() == "histogram":
            conf = self.marks_bar(conf, state, response)
        else:
            conf = self.marks_bar(conf, state, response)
        return conf

    def marks_pie(self, conf, state, response):
        donut = state._state["params"].get("isDonut")
        conf["marks"] = [
            {
                "type": "arc",
                "from": {"data": "table"},
                "encode": {
                    "enter": {
                        "fill": {"scale": "groupcolor", "field": "group"},
                        "x": {"signal": "width / 2"},
                        "y": {"signal": "height / 2"},
                        "startAngle": {"field": "startAngle"},
                        "endAngle": {"field": "endAngle"},
                        "innerRadius": {"signal": "width * .35"}
                        if donut
                        else {"value": 0},
                        "outerRadius": {"signal": "width / 2"},
                        **(
                            {"tooltip": {"field": "tooltip"}}
                            if state._state["params"]["addTooltip"]
                            else {}
                        ),
                    }
                },
            }
        ]
        if state._state["params"]["labels"]["show"]:
            conf["marks"].append(
                {
                    "type": "text",
                    "from": {"data": "table"},
                    "encode": {
                        "enter": {
                            "x": {"field": {"group": "width"}, "mult": 0.5},
                            "y": {"field": {"group": "height"}, "mult": 0.5},
                            "radius": {"signal": "width / 2", "offset": 8},
                            "theta": {
                                "signal": "(datum.startAngle + datum.endAngle)/2"
                            },
                            "fill": {"value": "#000"},
                            # Hide labels for small angles
                            "fillOpacity": {
                                "signal": "datum.endAngle - datum.startAngle < .1 ? 0 : 1"
                            },
                            "align": {
                                "signal": "(datum.startAngle + datum.endAngle)/2 < 3.14 ? 'left' : 'right'"
                            },
                            "baseline": {"value": "middle"},
                            "text": {"field": "label"},
                        }
                    },
                }
            )
        return conf

    def _marks_histogram(self, state, ax):
        ret = []
        x2 = (
            "axis"
            if state.multi_axis()
            else "group"
            if state.groups_side_by_side(ax)
            else "metric"
            if state.metrics_side_by_side(ax)
            else "x"
        )
        ret = [
            {
                "type": "group",
                "from": {"facet": {"data": "table", "name": "facet", "groupby": "x"}},
                "signals": [{"name": "width", "update": "bandwidth('xscale')"}],
                "encode": {"enter": {"x": {"scale": "xscale", "field": "x"}}},
                "scales": [
                    {
                        "name": "xscale2",
                        "type": "band",
                        "range": "width",
                        "domain": {"data": "facet", "field": x2},
                    }
                ],
                "marks": [],
            }
        ]
        marks = ret[0]["marks"]

        fill = (
            {"scale": "groupcolor", "field": "group"}
            if state.group_aggs()
            else {"scale": "metriccolor", "field": "metric"}
        )
        stacked = state.stacked_applied(ax)
        marks.append(
            {
                "type": "rect",
                "from": {"data": "facet"},
                "encode": {
                    "enter": {
                        "x": {"scale": "xscale2", "field": x2},
                        "width": {"scale": "xscale2", "band": 1},
                        "y": {
                            "scale": ax["id"],
                            "field": state.y(ax) + "|0" if stacked else state.y(ax),
                        },
                        "y2": {
                            "scale": ax["id"],
                            **(
                                {"field": state.y(ax) + "|1"}
                                if stacked
                                else {"value": 0}
                            ),
                        },
                        "fill": fill,
                        "fillOpacity": {"value": 0.8},
                        **(
                            {"tooltip": {"field": "tooltip"}}
                            if state._state["params"]["addTooltip"]
                            else {}
                        ),
                    }
                },
            }
        )
        return ret

    def _marks_line(self, state, ax):
        circlesizes = []
        strokesizes = []
        marks = []
        for agg in state.metric_aggs():
            params = state.series_params(agg)
            if params["valueAxis"] != ax["id"]:
                continue
            label = params["data"]["label"]
            size = params.get("lineWidth") or 4
            circletest = (
                f"datum.metric == '{label}'" if params["showCircles"] else "false"
            )
            circlesizes.append({"test": circletest, "value": (size * 2) ** 2})

            stroketest = (
                f"datum.metric == '{label}'"
                if params["drawLinesBetweenPoints"]
                else "false"
            )
            strokesizes.append({"test": stroketest, "value": size})
        circlesizes.append({"value": 0})
        strokesizes.append({"value": 0})

        stackgroupfield = "group" if state.group_aggs() else "metric"
        stylescale = "groupcolor" if state.group_aggs() else "metriccolor"

        marks.append(
            {
                "type": "line",
                "from": {"data": "facet"},
                "encode": {
                    "enter": {
                        "x": {"scale": "xscale", "field": "x"},
                        "y": {
                            "scale": ax["id"],
                            "field": state.y(ax) + "|1"
                            if state.stacked_applied(ax)
                            else state.y(ax),
                        },
                        "stroke": {"scale": stylescale, "field": stackgroupfield},
                        "strokeWidth": strokesizes,
                    }
                },
            }
        )
        marks.append(
            {
                "type": "symbol",
                "from": {"data": "facet"},
                "encode": {
                    "enter": {
                        "x": {"scale": "xscale", "field": "x"},
                        "y": {
                            "scale": ax["id"],
                            "field": state.y(ax) + "|1"
                            if state.stacked_applied(ax)
                            else state.y(ax),
                        },
                        "fill": {"scale": stylescale, "field": stackgroupfield},
                        "size": circlesizes,
                        **(
                            {"tooltip": {"field": "tooltip"}}
                            if state._state["params"]["addTooltip"]
                            else {}
                        ),
                    }
                },
            }
        )
        marks = [
            {
                "type": "group",
                "from": {
                    "facet": {
                        "data": "table",
                        "name": "facet",
                        "groupby": stackgroupfield,
                    }
                },
                "marks": marks,
            }
        ]
        return marks

    def marks_bar(self, conf, state, response):
        conf["marks"] = []
        ax = state.valueaxes()[0]
        marks = []

        for ax in state.valueaxes():
            handler = {"histogram": self._marks_histogram, "line": self._marks_line}[
                state.valueaxtype(ax)
            ]
            marks.extend(handler(state, ax))

        conf["marks"] = marks
        return conf

    def translate_legacy(self, visualization, response, scope):
        state = ContextVisualization(visualization=visualization, config=scope.config)

        ret = self.conf(state)
        ret = self.data(ret, state, response)
        ret = self.scales(ret, state)
        ret = self.axes(ret, state)
        ret = self.legends(ret, state)
        ret = self.marks(ret, state, response)
        return ret

    def translate_vega(self, visualization, response, scope):
        ret = hjson.loads(visualization.visState["params"]["spec"])
        data = ret["data"] if isinstance(ret["data"], dict) else ret["data"][0]
        data.pop("url")
        data["values"] = response.to_dict()
        ret.setdefault("width", DEFAULT_WIDTH)
        ret.setdefault("height", DEFAULT_HEIGHT)
        ret.setdefault("padding", DEFAULT_PADDING)
        return ret

    def translate(self, visualization, response, scope):
        """
        Transform a kibana visualization object and an elasticsearch_dsl response into a vega object.

        :param elasticsearch_dsl.Document visualization: Visualization fetched from a kibana index.
        :param elasticsearch_dsl.response.Response visualization: Visualization fetched from a kibana index.
        :param Scope scope: The scope associated for data fetching.
        """
        if visualization.visState["type"] == "vega":
            return self.translate_vega(visualization, response, scope)
        else:
            return self.translate_legacy(visualization, response, scope)
